/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor.fx.visual;

import static org.fortiss.tooling.base.ui.utils.LWFXEditorUtils.stickyConnectorHasVisibleConnections;

import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramLayers;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.widgets.AlignedText;
import org.fortiss.tooling.kernel.model.INamedElement;

import javafx.geometry.Pos;
import javafx.geometry.Rectangle2D;

/**
 * {@link IVisual} for layouted and named {@link IConnector}s of some content element within a
 * diagram (sticky connectors).
 * 
 * @author diewald
 */
public class NamedLayoutedCircularAnchorageContentVisual<T extends ILayoutedModelElement & INamedElement & IConnector>
		extends LayoutedCircularAnchorageContentVisualBase<T> {
	/** The name text label. */
	private final AlignedText nameText = new AlignedText();

	/** Constructor. */
	public NamedLayoutedCircularAnchorageContentVisual(IContentAnchorageMVCBundle mvcb,
			Class<T> modelType) {
		super(mvcb, modelType);
		nameText.setMouseTransparent(true);
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);
		if(enableName()) {
			nameText.setText(getName());
			updateNameTextPosition();
			if(nameText.getParent() == null) {
				layers.getVisualFeedbackLayer().add(nameText, getMVCBundle());
			}
		} else if(nameText.getParent() != null) {
			layers.getVisualFeedbackLayer().remove(nameText);
		}
	}

	/**
	 * Sets the position of the label text.
	 * Depending on the side, the text is placed off to the side.
	 */
	private void updateNameTextPosition() {
		Rectangle2D bounds = getCurrentBounds();
		switch(getSide()) {
			case LEFT:
				if(stickyConnectorHasVisibleConnections(getModelElement())) {
					// move the text a little to the side if there is a connection
					nameText.setX(bounds.getMinX() + bounds.getWidth() / 4);
					nameText.setY(bounds.getMinY() + bounds.getHeight() / 4);
					nameText.setAlignment(Pos.BOTTOM_RIGHT);
				} else {
					nameText.setX(bounds.getMinX());
					nameText.setY(bounds.getMinY() + bounds.getHeight() / 2);
					nameText.setAlignment(Pos.CENTER_RIGHT);
				}
				break;
			case RIGHT:
				if(stickyConnectorHasVisibleConnections(getModelElement())) {
					// move the text a little to the side if there is a connection
					nameText.setX(bounds.getMaxX() - bounds.getWidth() / 4);
					nameText.setY(bounds.getMinY() + bounds.getHeight() / 4);
					nameText.setAlignment(Pos.BOTTOM_LEFT);
				} else {
					nameText.setX(bounds.getMaxX());
					nameText.setY(bounds.getMinY() + bounds.getHeight() / 2);
					nameText.setAlignment(Pos.CENTER_LEFT);
				}
				break;
			case TOP:
				nameText.setX(bounds.getMaxX());
				nameText.setY(bounds.getMinY());
				nameText.setAlignment(Pos.BOTTOM_CENTER);
				break;
			case BOTTOM:
				nameText.setX(bounds.getMaxX());
				nameText.setY(bounds.getMaxY());
				nameText.setAlignment(Pos.TOP_CENTER);
				break;
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeAllVisuals(DiagramLayers layers) {
		super.removeAllVisuals(layers);
		if(nameText.getParent() != null) {
			layers.getVisualFeedbackLayer().remove(nameText);
		}
	}

	/** Returns whether the name label should be enabled. */
	public boolean enableName() {
		return !stickyConnectorHasVisibleConnections(getModelElement());
	}

	/** Returns the name. */
	public String getName() {
		return getModelElement().getName();
	}
}
