/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.layout.auto;

import static org.eclipse.elk.graph.util.ElkGraphUtil.createGraph;
import static org.eclipse.elk.graph.util.ElkGraphUtil.createLabel;
import static org.eclipse.elk.graph.util.ElkGraphUtil.createNode;
import static org.eclipse.elk.graph.util.ElkGraphUtil.createPort;
import static org.eclipse.elk.graph.util.ElkGraphUtil.createSimpleEdge;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_CONNECTOR_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_SIZE;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.addBendPointToConnection;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setNodeLayoutData;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setNodePosition;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setStickyConnectorLayoutData;

import java.util.Map.Entry;

import org.eclipse.elk.alg.layered.LayeredLayoutProvider;
import org.eclipse.elk.alg.layered.options.InteractiveReferencePoint;
import org.eclipse.elk.alg.layered.options.LayeredMetaDataProvider;
import org.eclipse.elk.alg.layered.options.LayeredOptions;
import org.eclipse.elk.core.AbstractLayoutProvider;
import org.eclipse.elk.core.data.LayoutMetaDataService;
import org.eclipse.elk.core.options.Direction;
import org.eclipse.elk.core.options.EdgeRouting;
import org.eclipse.elk.core.options.PortLabelPlacement;
import org.eclipse.elk.core.options.PortSide;
import org.eclipse.elk.core.util.BasicProgressMonitor;
import org.eclipse.elk.graph.ElkEdge;
import org.eclipse.elk.graph.ElkEdgeSection;
import org.eclipse.elk.graph.ElkLabel;
import org.eclipse.elk.graph.ElkNode;
import org.eclipse.elk.graph.ElkPort;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.BasicEMap;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.EMap;
import org.fortiss.tooling.base.model.base.EntryConnectorBase;
import org.fortiss.tooling.base.model.base.ExitConnectorBase;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Points;
import org.fortiss.tooling.base.model.layout.impl.PointsImpl;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * <p>
 * {@link IAutoLayouter} based on the Eclipse Layout Kernel (ELK).
 * </p>
 * <P>
 * Approach:
 * <ul>
 * <li>Generate a ELK graph from the input model</li>
 * <li>Apply the ELK <i>layered layout algorithm</i></li>
 * <li>Transfer generated layout to input model
 * </ul>
 * 
 * @author offtermatt, barner
 */
public class KielerAutoLayouter implements IAutoLayouter {

	/** Mapping from {@link IHierarchicElement}s to {@link ElkNode}s. */
	private EMap<IHierarchicElement, ElkNode> modelElementsToElkNodes;

	/** Mapping from outbound {@link IConnector}s to {@link ElkPort}s. */
	private EMap<IConnector, ElkPort> outboundConnectorsToElkPorts;

	/** Mapping from inbound {@link IConnector}s to {@link ElkPort}s. */
	private EMap<IConnector, ElkPort> inboundConnectorsToElkPorts;

	/** Used to map undirected {@link IConnector}s to {@link ElkPort}s */
	private EMap<IConnector, ElkPort> undirectedConnectorsToElkPorts;

	/** Mapping from {@link IConnection}s to {@link ElkEdge}s. */
	private EMap<IConnection, ElkEdge> connectionsToElkEdges;

	/** Mapping from global inbound {@link IConnector}s to {@link ElkPort}s. */
	private EMap<IConnector, ElkPort> globalInboundConnectorsToElkPorts;

	/** Mapping from global outbound {@link IConnector}s to {@link ElkPort}s. */
	private EMap<IConnector, ElkPort> globalOutboundConnectorsToElkPorts;

	/** {@inheritDoc} */
	@Override
	public void performAutoLayout(IHierarchicElement element) {

		LayoutMetaDataService.getInstance()
				.registerLayoutMetaDataProviders(new LayeredMetaDataProvider());

		ElkNode rootNode = createELKGraph(element);

		rootNode.setProperty(LayeredOptions.DIRECTION, Direction.RIGHT);
		rootNode.setProperty(LayeredOptions.INTERACTIVE_LAYOUT, true);
		rootNode.setProperty(LayeredOptions.FEEDBACK_EDGES, true);
		rootNode.setProperty(LayeredOptions.EDGE_ROUTING, EdgeRouting.ORTHOGONAL);
		rootNode.setProperty(LayeredOptions.SPACING_NODE_NODE_BETWEEN_LAYERS,
				5.0 * DEFAULT_GRID_SIZE);
		rootNode.setProperty(LayeredOptions.SPACING_EDGE_NODE_BETWEEN_LAYERS,
				2.0 * DEFAULT_GRID_SIZE);
		rootNode.setProperty(LayeredOptions.INTERACTIVE_REFERENCE_POINT,
				InteractiveReferencePoint.TOP_LEFT);
		rootNode.setProperty(LayeredOptions.PORT_LABELS_PLACEMENT, PortLabelPlacement.outside());

		AbstractLayoutProvider layoutProvider = new LayeredLayoutProvider();
		BasicProgressMonitor progressMonitor = new BasicProgressMonitor();
		layoutProvider.layout(rootNode, progressMonitor);

		applyLayout(rootNode, element);
	}

	/**
	 * Truncates a given {@code float} value and decreases the value until it is a multiple of the
	 * grid size.
	 * 
	 * @param f
	 *            {@code float} to process
	 * 
	 * @return Given float value, truncated and aligned to the underlying grid.
	 */
	private int truncateSnap2Grid(double f) {
		return truncateSnap2Grid(f, false);
	}

	/**
	 * Truncates a given {@code float} value and aligns it the grid size.
	 * 
	 * @param f
	 *            {@code float} to process
	 * 
	 * @param enlarge
	 *            Whether to enlarge the given value while aligning to the grid.
	 * 
	 * @return Given float value, truncated and aligned to the underlying grid.
	 */
	private int truncateSnap2Grid(double f, boolean enlarge) {
		int rval = (int)f;

		final int deviationFromGrid = rval % DEFAULT_GRID_SIZE;
		if(deviationFromGrid > 0) {
			if(enlarge) {
				rval += DEFAULT_GRID_SIZE;
			}
		}
		rval -= deviationFromGrid;
		return rval;
	}

	/**
	 * Applies the layout defined in the {@link ElkNode} follows to given
	 * {@link IHierarchicElement}.
	 * 
	 * @param template
	 *            The {@link ElkNode} that stores the layout.
	 * @param element
	 *            The {@link IHierarchicElement} to which the layout should be applied.
	 */
	private void applyLayout(ElkNode template, IHierarchicElement element) {
		ITopLevelElement topLevelElement =
				IPersistencyService.getInstance().getTopLevelElementFor(element);
		topLevelElement.runAsCommand(() -> {
			filterBendPoints(element);

			for(IHierarchicElement child : element.getContainedElements()) {
				ElkNode elkNode = modelElementsToElkNodes.get(child);
				setNodeLayoutData((ILayoutedModelElement)child, truncateSnap2Grid(elkNode.getX()),
						truncateSnap2Grid(elkNode.getY()),
						truncateSnap2Grid(elkNode.getWidth(), true),
						truncateSnap2Grid(elkNode.getHeight(), true));
			}

			for(Entry<IConnection, ElkEdge> entry : connectionsToElkEdges) {
				ElkEdge elkEdge = entry.getValue();
				IConnection connection = entry.getKey();
				for(ElkEdgeSection section : elkEdge.getSections()) {
					for(var bendPoint : section.getBendPoints()) {
						addBendPointToConnection((ILayoutedModelElement)connection,
								truncateSnap2Grid(bendPoint.getX()),
								truncateSnap2Grid(bendPoint.getY()));
					}
				}
			}

			setPositionsForConnectors(outboundConnectorsToElkPorts);
			setPositionsForConnectors(inboundConnectorsToElkPorts);
			setPositionsForGlobalConnectors(globalInboundConnectorsToElkPorts);
			setPositionsForGlobalConnectors(globalOutboundConnectorsToElkPorts);
		});
	}

	/**
	 * Applies the layout of the specified {@link ElkPort}s to their matching (global)
	 * {@link IConnector}s. Global means in this case that the {@link IConnector}s do not have a
	 * corresponding node, i.e. because they contain connections to a higher-level model.
	 * 
	 * @param globalConnectors
	 *            A bijection from {@link IConnector}s to {@link ElkPort}s
	 */
	private void setPositionsForGlobalConnectors(EMap<IConnector, ElkPort> globalConnectors) {
		for(Entry<IConnector, ElkPort> entry : globalConnectors) {
			IConnector connector = entry.getKey();
			ElkPort elkPort = entry.getValue();
			ElkNode tempNode = elkPort.getParent();
			setNodePosition((ILayoutedModelElement)connector, truncateSnap2Grid(tempNode.getX()),
					truncateSnap2Grid(tempNode.getY()));
		}
	}

	/**
	 * Applies the layout of the specified {@link ElkPort}s to their matching {@link IConnector}s.
	 * The
	 * {@link IConnector}s must have corresponding nodes, i.e. they can not contain an
	 * {@link IConnection} to a higher-level model.
	 * 
	 * @param connectors
	 *            A bijection from {@link IConnector}s to {@link ElkPort}s.
	 */
	private void setPositionsForConnectors(EMap<IConnector, ElkPort> connectors) {
		for(Entry<IConnector, ElkPort> entry : connectors) {
			ElkPort elkPort = entry.getValue();
			IConnector connector = entry.getKey();
			PortSide side = elkPort.getProperty(LayeredOptions.PORT_SIDE);
			EOrientation orientation = null;
			switch(side) {
				case NORTH:
					orientation = EOrientation.NORTH;
					break;
				case WEST:
					orientation = EOrientation.WEST;
					break;
				case EAST:
					orientation = EOrientation.EAST;
					break;
				case SOUTH:
					orientation = EOrientation.SOUTH;
					break;
			}

			// The parent component need not be aligned to the grid. The Y-coordinate
			// of the parent is truncated. Since elkPort.getY() is relative to the
			// Y-coordinate of the parent, we need to correct when snapping to the grid.
			double parentGridOffset = elkPort.getParent().getY() % DEFAULT_GRID_SIZE;
			setStickyConnectorLayoutData((ILayoutedModelElement)connector, orientation,
					truncateSnap2Grid(elkPort.getY() + parentGridOffset));
		}
	}

	/** Removes all bend-points in connections contained in the given {@link IHierarchicElement}. */
	private void filterBendPoints(IHierarchicElement element) {
		for(IConnection connection : element.getConnections()) {
			for(ILayoutData layoutData : ((ILayoutedModelElement)connection).getLayoutData()) {
				if(layoutData instanceof Points || layoutData instanceof PointsImpl) {
					((Points)layoutData).getPoints().clear();
				}
			}
		}
	}

	/**
	 * Creates a KIELER {@link ElkNode} graph with the same structure as the given
	 * {@link IHierarchicElement}.
	 * 
	 * @param element
	 *            The {@link IHierarchicElement} for which the {@link ElkNode} graph should be
	 *            created
	 * 
	 * @return A {@link ElkNode} graph with the same structure as the graph contained in the
	 *         {@link IHierarchicElement}.
	 */
	private ElkNode createELKGraph(IHierarchicElement element) {

		ElkNode rootNode = createGraph();

		modelElementsToElkNodes = new BasicEMap<IHierarchicElement, ElkNode>();
		outboundConnectorsToElkPorts = new BasicEMap<IConnector, ElkPort>();
		inboundConnectorsToElkPorts = new BasicEMap<IConnector, ElkPort>();
		connectionsToElkEdges = new BasicEMap<IConnection, ElkEdge>();
		undirectedConnectorsToElkPorts = new BasicEMap<IConnector, ElkPort>();

		IAutoLayouterTapeMeasure tapeMeasure = new DiagramTapeMeasure();

		// Create nodes
		for(IHierarchicElement child : element.getContainedElements()) {
			if(!(child instanceof ILayoutedModelElement)) {
				continue;
			}
			// Adjust size of node to number of connectors and label size
			ElkNode elkNode = createNode(rootNode);

			int width = tapeMeasure.getElementWidth(child);
			elkNode.setWidth(truncateSnap2Grid(width, true));
			int height = tapeMeasure.getElementHeight(child);
			elkNode.setHeight(truncateSnap2Grid(height, true));
			modelElementsToElkNodes.put(child, elkNode);

			// Setup connectors
			EList<IConnector> inputConnectors = new BasicEList<IConnector>();
			EList<IConnector> outputConnectors = new BasicEList<IConnector>();
			EList<IConnector> undirectedConnectors = new BasicEList<IConnector>();
			for(IConnector connector : child.getConnectors()) {
				if(connector instanceof EntryConnectorBase) {
					inputConnectors.add(connector);
				} else if(connector instanceof ExitConnectorBase) {
					outputConnectors.add(connector);
				} else {
					outputConnectors.add(connector);
					undirectedConnectors.add(connector);
				}
			}
			for(IConnector connector : outputConnectors) {
				ElkPort k = createKPortFromIConnector(tapeMeasure, connector, elkNode,
						outboundConnectorsToElkPorts);
				if(undirectedConnectors.contains(connector)) {
					undirectedConnectorsToElkPorts.put(connector, k);
				}
			}
			for(IConnector connector : inputConnectors) {
				createKPortFromIConnector(tapeMeasure, connector, elkNode,
						inboundConnectorsToElkPorts);
			}
		}

		// Create virtual nodes for global ports
		globalInboundConnectorsToElkPorts = new BasicEMap<IConnector, ElkPort>();
		globalOutboundConnectorsToElkPorts = new BasicEMap<IConnector, ElkPort>();

		for(IConnector connector : element.getConnectors()) {
			ElkNode elkNodeVirtual = createNode(rootNode);
			elkNodeVirtual.setWidth(tapeMeasure.getConnectorWidth(connector));
			elkNodeVirtual.setHeight(tapeMeasure.getConnectorHeight(connector));

			ElkPort elkPort = createPort(elkNodeVirtual);
			if(connector instanceof ExitConnectorBase) {
				globalOutboundConnectorsToElkPorts.put(connector, elkPort);
			} else {
				globalInboundConnectorsToElkPorts.put(connector, elkPort);
				outboundConnectorsToElkPorts.put(connector, elkPort);
			}
		}

		// Create edges
		for(Entry<IConnector, ElkPort> entry : outboundConnectorsToElkPorts) {
			for(IConnection connection : entry.getKey().getOutgoing()) {
				ElkPort kSourcePort = entry.getValue();

				ElkPort kTargetPort = inboundConnectorsToElkPorts.get(connection.getTarget());
				if(kTargetPort == null) {
					kTargetPort = globalOutboundConnectorsToElkPorts.get(connection.getTarget());
				}
				if(kTargetPort == null) {
					kTargetPort = undirectedConnectorsToElkPorts.get(connection.getTarget());
				}

				ElkEdge elkEdge = createSimpleEdge(kSourcePort, kTargetPort);

				ElkLabel elkLabel = createLabel(elkEdge);
				int width = tapeMeasure.getConnectionWidth(connection);
				elkLabel.setWidth(truncateSnap2Grid(width, true));

				connectionsToElkEdges.put(connection, elkEdge);
			}
		}

		return rootNode;
	}

	/**
	 * Converts a given {@link IConnector} into a {@link ElkPort}, and registers it with the given
	 * {@link ElkNode} (representing the {@link IHierarchicElement} containing the
	 * {@link IConnector})
	 * and a {@link IConnector}-to- {@link ElkPort} map.
	 * 
	 * @param tapeMeasure
	 *            {@link IAutoLayouterTapeMeasure} to estimate the label width
	 * @param connector
	 *            {@link IConnector} to be converted to {@link ElkPort}
	 * @param elkNode
	 *            {@link ElkNode} that will contain the newly created {@link ElkPort}
	 * @param connectorsToelkPorts
	 *            Map between {@link IConnector}s and {@link ElkPort}s.
	 * 
	 * @return {@link ElkNode} representing the given {@link IConnector}.
	 */
	private ElkPort createKPortFromIConnector(IAutoLayouterTapeMeasure tapeMeasure,
			IConnector connector, ElkNode elkNode, EMap<IConnector, ElkPort> connectorsToelkPorts) {
		ElkPort elkPort = createPort(elkNode);
		elkPort.setParent(elkNode);
		elkPort.setHeight(DEFAULT_CONNECTOR_SIZE);

		if(connector instanceof INamedElement) {
			ElkLabel elkPortLabel = createLabel(elkPort);
			String name = ((INamedElement)connector).getName();
			elkPortLabel.setWidth(tapeMeasure.getConnectorLabelWidth(connector));
			elkPortLabel.setText(name);
		}

		connectorsToelkPorts.put(connector, elkPort);
		return elkPort;
	}
}
